import sounddevice

from ._version import __version__


CREDITS = (
    'Eternal Radio Player\n'
    'License: GPLv3+ - https://github.com/alexitx/eternal-radio-player/blob/master/LICENSE\n'
    '---\n'
    'numpy - https://github.com/numpy/numpy\n'
    'License: BSD 3-Clause - https://github.com/numpy/numpy/blob/main/LICENSE.txt\n'
    '---\n'
    'platformdirs - https://github.com/platformdirs/platformdirs\n'
    'License: MIT - https://github.com/platformdirs/platformdirs/blob/main/LICENSE.txt\n'
    '---\n'
    'pyminiaudio - https://github.com/irmen/pyminiaudio\n'
    'License: MIT - https://github.com/irmen/pyminiaudio/blob/master/LICENSE\n'
    '---\n'
    'python-sounddevice - https://github.com/spatialaudio/python-sounddevice\n'
    'License: MIT - https://github.com/spatialaudio/python-sounddevice/blob/master/LICENSE\n'
    '---\n'
    'requests - https://github.com/psf/requests\n'
    'License: Apache 2.0 - https://github.com/psf/requests/blob/main/LICENSE\n'
    '---\n'
    'schema - https://github.com/keleshev/schema\n'
    'License: MIT - https://github.com/keleshev/schema/blob/master/LICENSE-MIT\n'
    '---\n'
    'timeago - https://github.com/hustcc/timeago\n'
    'License: MIT - https://github.com/hustcc/timeago/blob/master/LICENSE\n'
    '---\n'
    'PyQtDarkTheme - https://github.com/5yutan5/PyQtDarkTheme\n'
    'License: MIT - https://github.com/5yutan5/PyQtDarkTheme/blob/main/LICENSE.txt\n'
    '---\n'
    'PySide6 - https://wiki.qt.io/Qt_for_Python\n'
    'License: LGPLv3 - https://tldrlegal.com/license/gnu-lesser-general-public-license-v3-(lgpl-3)\n'
    '---\n'
    'RemixIcon - https://github.com/Remix-Design/RemixIcon\n'
    'License: Apache 2.0 - https://github.com/Remix-Design/RemixIcon/blob/master/License'
)

USER_AGENT = f'Eternal Radio Player/{__version__}'
REQUEST_TIMEOUT = 5.0

STREAM_URL = 'https://radio.jump.bg/proxy/mnikolov/stream'
STREAM_ITER_CHUNK_SIZE = 16 * 1024

PLAYER_FRAME_COUNT = 4 * 1024

# Default sounddevice host API
for index, host_api in enumerate(sounddevice.query_hostapis()):
    if host_api['name'] == 'Windows WASAPI' and host_api['default_output_device'] != -1:
        SOUNDDEVICE_HOST_API = index
        break
else:
    SOUNDDEVICE_HOST_API = sounddevice.default.hostapi

RECENT_SONGS_URL = 'https://radio.jump.bg/recentfeed/mnikolov/json'
RECENT_SONGS_CACHE_TIME = 10.0
RECENT_SONGS_UPDATE_TIME = 30.0

CONFIG_DEFAULTS = {
    'connection-timeout': REQUEST_TIMEOUT,
    'recent-songs-update-time': RECENT_SONGS_UPDATE_TIME,
    'volume': 1.0
}



__author__ = 'alexitx'
__author_email__ = 'alexander@alexitx.com'
__license__ = 'GPLv3+'
__title__ = 'eternal-radio-player'
__url__ = 'https://github.com/alexitx/eternal-radio-player'
__version__ = '0.1.1'



import logging
from PySide6 import QtCore


class _SignalProxy(QtCore.QObject):

    message = QtCore.Signal(str)


class QPlainTextEditHandler(logging.Handler):

    def __init__(self, widget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widget = widget
        self._signal_proxy = _SignalProxy()
        self.message = self._signal_proxy.message

    def emit(self, record):
        self._signal_proxy.message.emit(self.format(record))



import logging
import miniaudio
import requests

from .constants import REQUEST_TIMEOUT, STREAM_ITER_CHUNK_SIZE, STREAM_URL, USER_AGENT
from .exceptions import StreamError


log = logging.getLogger(__name__)


class HTTPStreamSource(miniaudio.StreamableSource):

    def __init__(self, request, iter_chunk_size=STREAM_ITER_CHUNK_SIZE):
        super().__init__()
        self._request = request
        self._stream_iter = self._request.raw.stream(iter_chunk_size, False)
        self._buffer = bytearray()

    def read(self, size):
        if self._request.raw.closed:
            raise StreamError('Cannot read from a closed stream')
        while len(self._buffer) < size:
            try:
                self._buffer.extend(next(self._stream_iter))
            except StopIteration:
                break
        data = self._buffer[:size]
        del self._buffer[:size]
        return data


def stream_request(timeout=REQUEST_TIMEOUT, **kwargs):
    log.debug(f"Making stream request with URL: '{STREAM_URL}', timeout: {timeout}")

    response = requests.get(
        url=STREAM_URL,
        headers={'User-Agent': USER_AGENT},
        timeout=timeout,
        stream=True,
        **kwargs
    )

    content_type = response.headers.get('Content-Type')
    if content_type == 'audio/mpeg':
        encoding = miniaudio.FileFormat.MP3
    elif content_type == 'audio/ogg':
        encoding = miniaudio.FileFormat.VORBIS
    else:
        raise ValueError(f'Unsupported stream encoding: {content_type}')

    icy_sr = response.headers.get('icy-sr')
    try:
        sample_rate = int(icy_sr)
    except ValueError:
        sample_rate = None

    return response, encoding, sample_rate



class EternalRadioPlayerError(Exception):
    pass


class ConfigError(EternalRadioPlayerError):
    pass


class StreamError(EternalRadioPlayerError):
    pass


class PlayerError(EternalRadioPlayerError):
    pass


class ResourceError(EternalRadioPlayerError):
    pass



from PySide6 import QtCore, QtGui, QtWidgets


class ControlToolButton(QtWidgets.QToolButton):

    hovered = QtCore.Signal(bool)
    focused = QtCore.Signal(bool)

    def enterEvent(self, event):
        self.hovered.emit(True)
        super().enterEvent(event)

    def leaveEvent(self, event):
        self.hovered.emit(False)
        super().leaveEvent(event)

    def focusInEvent(self, event):
        self.focused.emit(True)
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        self.focused.emit(False)
        super().focusOutEvent(event)


class RecentSong(QtWidgets.QWidget):

    def __init__(
        self,
        parent=None,
        title=None,
        timestamp=None,
        margins=8,
        spacing=6,
        title_font_size=12,
        timestamp_font_size=10,
        **kwargs
    ):
        super().__init__(parent=parent, **kwargs)
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(margins, margins, margins, margins)
        self._layout.setSpacing(spacing)
        self.title = QtWidgets.QLabel(self, text=title)
        self.title.setObjectName('title')
        self.title.setWordWrap(True)
        self.set_title_font_size(title_font_size)
        self.timestamp = QtWidgets.QLabel(self, text=timestamp)
        self.timestamp.setObjectName('timestamp')
        self.timestamp.setWordWrap(True)
        self.set_timestamp_font_size(timestamp_font_size)
        self._layout.addWidget(self.title)
        self._layout.addWidget(self.timestamp)
        self.setLayout(self._layout)

    def set_title_font_size(self, value):
        font = self.title.font()
        font.setPointSize(value)
        self.title.setFont(font)

    def set_timestamp_font_size(self, value):
        font = self.timestamp.font()
        font.setPointSize(value)
        self.timestamp.setFont(font)

    def paintEvent(self, pe):
        opt = QtWidgets.QStyleOption()
        opt.initFrom(self)
        painter = QtGui.QPainter(self)
        s = self.style()
        s.drawPrimitive(QtWidgets.QStyle.PE_Widget, opt, painter, self)



import logging
import platform
import traceback
import miniaudio
import sounddevice

from ._version import __version__
from .exceptions import ResourceError


log = logging.getLogger(__name__)


def format_exc(exc, tb=False):
    if str(exc):
        msg = f'{type(exc).__name__}: {exc}'
        if tb:
            return f'{msg}\n{traceback.format_exc()}'
        return msg
    return f'{type(exc).__name__}: {traceback.format_exc()}'


def os_version():
    if platform.system() == 'Windows':
        return f'{platform.platform()} {platform.machine()}'
    return platform.platform()


def python_version():
    return f'{platform.python_implementation()} {platform.python_version()} [{platform.python_compiler()}]'


def miniaudio_version():
    return miniaudio.__version__


def miniaudio_backend_version():
    return miniaudio.lib_version()


def sounddevice_version():
    return sounddevice.__version__


def sounddevice_backend_version():
    return sounddevice.get_portaudio_version()[1]


def pyside_version():
    try:
        import PySide6
    except ImportError:
        return None
    return PySide6.__version__


def qt_version():
    try:
        from PySide6.QtCore import qVersion
    except ImportError:
        return None
    return qVersion()


def system_info():
    return (
        f'Eternal Radio Player {__version__}\n'
        f'OS: {os_version()}\n'
        f'Python: {python_version()}\n'
        f'miniaudio: {miniaudio_version()}\n'
        f'miniaudio Backend: {miniaudio_backend_version()}\n'
        f'sounddevice: {sounddevice_version()}\n'
        f'sounddevice Backend: {sounddevice_backend_version()}\n'
        f'PySide: {pyside_version()}\n'
        f'Qt: {qt_version()}'
    )


def qt_resource(path, text=False):
    from PySide6 import QtCore

    file = QtCore.QFile(path)
    if text:
        is_open = file.open(QtCore.QIODevice.ReadOnly | QtCore.QIODevice.Text)
        file_wrapper = QtCore.QTextStream(file)
    else:
        is_open = file.open(QtCore.QIODevice.ReadOnly)
        file_wrapper = file
    if not is_open:
        raise ResourceError(f'{file.error()}: {file.errorString()}')
    data = file_wrapper.readAll()
    file.close()
    return data



import argparse
import cmd
import logging
import os.path
import platform
import sys
import platformdirs
import timeago
from pathlib import Path

from ._version import __version__
from .config import Config
from .constants import CREDITS
from .exceptions import PlayerError
from .player import get_output_device, get_output_devices, RadioPlayer
from .utils import format_exc, system_info


log = logging.getLogger(__name__)


class App(cmd.Cmd):

    prompt = '> '
    intro = (
        f"\nEternal Radio Player {__version__}\n"
        "Type 'help' for a list of commands or 'quit' to exit\n"
    )
    _hidden_attrs = ('do_EOF', 'do_help')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug('Initializing CLI command logger')
        logger = logging.getLogger('cli-cmd')
        logger.propagate = False
        logger.setLevel(logging.NOTSET)
        handler = logging.StreamHandler()
        logger.addHandler(handler)
        self._log = logger
        self._player = RadioPlayer(Config.data['output-device'], Config.data['connection-timeout'])
        self._player.set_volume(Config.data['volume'])

    def do_play(self, _):
        try:
            self._player.play()
        except PlayerError as e:
            log.error(f'Error starting player: {e}')

    def do_stop(self, _):
        self._player.stop()

    def do_volume(self, args):
        if not args:
            self._log.info('Not enough arguments')
            return
        try:
            volume = int(args)
        except ValueError:
            self._log.info(f"Invalid value '{args}'")
            return
        if not 0 <= volume <= 100:
            self._log.info('The value must be within range 0-100')
            return
        volume_float = volume / 100
        self._player.set_volume(volume_float)
        Config.data['volume'] = volume_float

    def do_recent(self, _):
        recent_songs_localized = []
        connection_timeout = Config.data['connection-timeout']
        for recent_song in self._player.get_recent_songs(connection_timeout):
            title = recent_song['title']
            timestamp = recent_song['timestamp']
            timestamp_fmt = timeago.format(timestamp)
            recent_songs_localized.append(f'{title}\n{timestamp_fmt}')
        self._log.info('\n\n'.join(recent_songs_localized))

    def do_config(self, args):
        if not args:
            self._log.info('Not enough arguments')
            return

        args = args.split(' ', 1)
        subcmd = args[0]
        value = args[1] if len(args) > 1 else None

        if subcmd == 'connection-timeout':
            if value is None:
                connection_timeout = Config.data['connection-timeout']
                self._log.info(f"Connection Timeout: {connection_timeout}s")
                return

            try:
                connection_timeout = float(value)
            except ValueError:
                self._log.info(f"Invalid value '{value}'")
                return
            if not 1.0 <= connection_timeout <= 60.0:
                self._log.info('The value be within range 1-60')
                return

            if self._player.running:
                self._player.stop()
            self._player.request_timeout = connection_timeout
            Config.data['connection-timeout'] = connection_timeout
            self._log.info(f'Connection timeout: {connection_timeout}s')
        elif subcmd == 'output-device':
            output_devices = get_output_devices()
            if value is None:
                output_devices_str = '\n'.join([f"#{device['index']} - {device['name']}" for device in output_devices])
                current_device_str = f"#{self._player.output_device['index']} - {self._player.output_device['name']}"
                self._log.info(
                    f'Output device: {current_device_str}\n\n'
                    'Available devices:\n'
                    f'{output_devices_str}'
                )
                return
            try:
                selected_device_index = int(value)
            except ValueError:
                self._log.info(f"Invalid value '{value}'")
                return
            try:
                device = get_output_device(selected_device_index)
            except PlayerError:
                self._log.info(f"Device with index '{value}' not found")
                return
            if self._player.running:
                self._player.stop()
            self._player.output_device = device
            Config.data['output-device'] = device['index']
            self._log.info(f"Output device: #{device['index']} - {device['name']}")
            return
        elif subcmd == 'save':
            Config.save()
        else:
            self._log.info(f"Invalid argument '{subcmd}'")

    def do_about(self, _):
        self._log.info(system_info())

    def do_credits(self, _):
        self._log.info(CREDITS)

    def do_quit(self, _):
        self._player.stop()
        Config.save()
        return True

    def do_help(self, args):
        if args:
            print()
        super().do_help(args)
        if args:
            print()

    def help_play(self):
        self._log.info(
            'Start the radio player\n\n'
            'Usage:\n'
            '  play'
        )

    def help_stop(self):
        self._log.info(
            'Stop the radio player\n\n'
            'Usage:\n'
            '  stop'
        )

    def help_volume(self):
        self._log.info(
            'View or change the sound volume\n\n'
            'Usage:\n'
            '  volume [0-100]'
        )

    def help_recent(self):
        self._log.info(
            'View the recently played songs\n\n'
            'Usage:\n'
            '  recent'
        )

    def help_config(self):
        self._log.info(
            'View or change the configuration\n\n'
            'Usage:\n'
            '  config connection-timeout [value]\n'
            '  config output-device [index]'
            '  config save'
        )

    def help_about(self):
        self._log.info(
            'View system information\n\n'
            'Usage:\n'
            '  about'
        )

    def help_credits(self):
        self._log.info(
            'View credits and licenses\n\n'
            'Usage:\n'
            '  credits'
        )

    def help_quit(self):
        self._log.info(
            'Quit the application\n\n'
            'Usage:\n'
            '  quit'
        )

    def get_names(self):
        return [n for n in dir(self.__class__) if n not in self._hidden_attrs]

    def cmdloop(self, *args, **kwargs) -> None:
        try:
            super().cmdloop(*args, **kwargs)
        except KeyboardInterrupt:
            log.debug('Received KeyboardInterrupt, stopping and exiting')
            self._player.stop()
            raise


def main():

    class HelpFormatter(argparse.HelpFormatter):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, max_help_position=32, **kwargs)

        def _format_action_invocation(self, action):
            if not action.option_strings or action.nargs == 0:
                return super()._format_action_invocation(action)
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            return f"{', '.join(action.option_strings)} {args_string}"

    parser = argparse.ArgumentParser(formatter_class=HelpFormatter)
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=f'Eternal Radio Player {__version__}'
    )
    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help='Set logging verbosity to debug'
    )
    parser.add_argument(
        '-c',
        '--cli',
        action='store_true',
        help='Run in CLI mode'
    )
    parser.add_argument(
        '-l',
        '--log',
        help='Log file path',
        metavar='<file>'
    )
    parser.add_argument(
        '-C',
        '--config',
        help='Config file path',
        metavar='<file>'
    )
    args = parser.parse_args()

    def error(msg, exc=None, tb=False):
        if exc:
            print(f'{msg}: {format_exc(exc, tb)}', file=sys.stderr)
        else:
            print(msg, file=sys.stderr)
        sys.exit(1)

    if args.log:
        log_dir = Path(os.path.dirname(args.log))
        log_file = Path(args.log)
    else:
        if platform.system() == 'Windows':
            dir_name = 'Eternal Radio Player'
        else:
            dir_name = 'eternal-radio-player'
        log_dir = platformdirs.user_log_path(dir_name, False)
        log_file = log_dir.joinpath('eternal-radio-player.log')
    try:
        log_dir.mkdir(parents=True, exist_ok=True)
    except OSError as e:
        error(f"Could not create log directory '{log_dir}'", e)

    root_logger = logging.getLogger()
    file_formatter = logging.Formatter('%(asctime)s %(levelname)s [%(name)s:%(funcName)s]: %(message)s')
    stream_formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(stream_formatter)
    stream_handler.setLevel(logging.NOTSET if args.debug else logging.INFO)
    file_handler = logging.FileHandler(str(log_file), 'w', 'utf-8')
    file_handler.setFormatter(file_formatter)
    root_logger.addHandler(stream_handler)
    root_logger.addHandler(file_handler)
    root_logger.setLevel(logging.NOTSET)

    log.debug(f'System information:\n{system_info()}')

    defaults = {
        'output-device': get_output_device()['index']
    }
    Config.init(args.config)
    Config.load(defaults)

    if not args.cli:
        try:
            import PySide6
            import qdarktheme
        except ImportError:
            log.debug('GUI dependencies not available')
        else:
            log.info('Starting in GUI mode')
            from .gui.gui import gui_main
            sys.exit(gui_main())

    try:
        log.info('Starting in CLI mode')
        App().cmdloop()
        log.info('Exiting')
        sys.exit()
    except KeyboardInterrupt:
        print('Interrupted')
        sys.exit(1)



import logging
import time
import timeago
import qdarktheme
from functools import partial
from PySide6 import QtCore, QtGui, QtWidgets

from ..config import Config
from ..constants import CREDITS, RECENT_SONGS_UPDATE_TIME, REQUEST_TIMEOUT
from ..exceptions import PlayerError
from ..player import RadioPlayer
from ..utils import qt_resource, system_info
from .generated.main_window import Ui_MainWindow
from .logging_ import QPlainTextEditHandler
from .widgets import RecentSong


log = logging.getLogger(__name__)


class RecentSongsUpdateWorker(QtCore.QObject):

    result = QtCore.Signal(list)

    def __init__(self, update_time=RECENT_SONGS_UPDATE_TIME, timeout=REQUEST_TIMEOUT, **kwargs):
        super().__init__(**kwargs)
        self.update_time = update_time
        self.timeout = timeout
        self._running = False

    def run(self):
        self._running = True
        try:
            data = RadioPlayer.get_recent_songs(self.timeout)
        except PlayerError as e:
            log.error(f'Recent songs update worker error: {e}')
        else:
            self.result.emit(data)
        last_update = time.time()
        while self._running:
            if time.time() - last_update < self.update_time:
                time.sleep(1)
                continue
            try:
                data = RadioPlayer.get_recent_songs(self.timeout)
            except PlayerError as e:
                log.error(f'Recent songs update worker error: {e}')
            else:
                self.result.emit(data)
            last_update = time.time()

    def stop(self):
        self._running = False


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, *args, logging_level=logging.INFO, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Set up logging handler for the console
        log.debug('Initializing GUI console logging handler')
        root_logger = logging.getLogger()
        self.console_handler = QPlainTextEditHandler(self.ui.console, level=logging_level)
        formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s', '%H:%M:%S')
        self.console_handler.setFormatter(formatter)
        self.console_handler.message.connect(self.ui.console.appendPlainText)
        root_logger.addHandler(self.console_handler)

        self.player = RadioPlayer()
        self.player.set_volume(Config.data['volume'])
        self.playing = False
        self.muted = False

        self.play_icon_normal = QtGui.QIcon(':icons/play.svg')
        self.play_icon_accent = QtGui.QIcon(':icons/play_accent.svg')
        self.stop_icon_normal = QtGui.QIcon(':icons/stop.svg')
        self.stop_icon_accent = QtGui.QIcon(':icons/stop_accent.svg')
        self.volume_normal_icon_normal = QtGui.QIcon(':icons/volume-normal.svg')
        self.volume_normal_icon_accent = QtGui.QIcon(':icons/volume-normal_accent.svg')
        self.volume_muted_icon_normal = QtGui.QIcon(':icons/volume-muted.svg')
        self.volume_muted_icon_accent = QtGui.QIcon(':icons/volume-muted_accent.svg')
        self.output_device_icon_normal = QtGui.QIcon(':icons/output-device.svg')
        self.output_device_icon_accent = QtGui.QIcon(':icons/output-device_accent.svg')
        self.console_icon_normal = QtGui.QIcon(':icons/console.svg')
        self.console_icon_accent = QtGui.QIcon(':icons/console_accent.svg')
        self.settings_icon_normal = QtGui.QIcon(':icons/settings.svg')
        self.settings_icon_accent = QtGui.QIcon(':icons/settings_accent.svg')

        # Assign generic handlers for control button events
        control_buttons_icons = (
            (self.ui.output_device_button, self.output_device_icon_normal, self.output_device_icon_accent),
            (self.ui.console_button, self.console_icon_normal, self.console_icon_accent),
            (self.ui.settings_button, self.settings_icon_normal, self.settings_icon_accent)
        )
        for widget, icon_normal, icon_accent in control_buttons_icons:
            fn = partial(self.set_widget_icon, widget, icon_normal, icon_accent)
            widget.hovered.connect(fn)
            widget.focused.connect(fn)

        # Assign special handlers for play/stop and mute/unmute button events
        self.ui.play_button.hovered.connect(self.set_play_button_icon)
        self.ui.play_button.focused.connect(self.set_play_button_icon)
        self.ui.play_button.toggled.connect(self.on_play)
        self.ui.volume_button.hovered.connect(self.set_volume_button_icon)
        self.ui.volume_button.focused.connect(self.set_volume_button_icon)
        self.ui.volume_button.clicked.connect(self.on_mute)

        self.ui.volume_slider.setValue(int(Config.data['volume'] * 100))
        self.ui.volume_slider.valueChanged.connect(self.on_volume_change)

        self.ui.console_button.clicked.connect(lambda: self.ui.main_widget.setCurrentIndex(1))
        self.ui.settings_button.clicked.connect(lambda: self.ui.main_widget.setCurrentIndex(2))

        # Setup console page widgets
        console_font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        console_font.setPointSize(8)
        self.ui.console.setFont(console_font)
        self.ui.console_back_button.clicked.connect(lambda: self.ui.main_widget.setCurrentIndex(0))

        # Setup settings page widgets
        settings_infos_font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        settings_infos_font.setPointSize(9)
        self.ui.system_info.setFont(settings_infos_font)
        self.ui.credits.setFont(settings_infos_font)
        self.ui.system_info.setPlainText(system_info())
        self.ui.credits.setPlainText(CREDITS)
        self.ui.connection_timeout_input.setValue(Config.data['connection-timeout'])
        self.ui.recent_songs_update_time_input.setValue(Config.data['recent-songs-update-time'])
        self.ui.settings_save_button.clicked.connect(self.save_settings)
        self.ui.settings_back_button.clicked.connect(lambda: self.ui.main_widget.setCurrentIndex(0))

        # Setup recent songs page widgets
        self.recent_songs_container_layout = QtWidgets.QVBoxLayout(self.ui.recent_songs_container)
        self.recent_songs_container_layout.setContentsMargins(0, 0, 0, 0)
        self.recent_songs_spacer = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)

        # Setup recent songs update worker
        self.recent_songs_widgets = []
        self.recent_songs_update_worker_thread = QtCore.QThread(self)
        self.recent_songs_update_worker = RecentSongsUpdateWorker(Config.data['recent-songs-update-time'])
        self.recent_songs_update_worker.result.connect(self.update_recent_songs)
        self.recent_songs_update_worker.moveToThread(self.recent_songs_update_worker_thread)
        self.recent_songs_update_worker_thread.started.connect(self.recent_songs_update_worker.run)

        log.info('GUI initialized')

        log.debug('Starting recent songs update worker')
        self.recent_songs_update_worker_thread.start()

    def on_play(self, state):
        self.playing = state
        self.set_play_button_icon(self.ui.play_button.underMouse() or self.ui.play_button.hasFocus())
        self.ui.play_button.repaint()
        if self.playing:
            try:
                self.player.play()
            except PlayerError as e:
                log.error(f'Could not start player: {e}')
                self.ui.play_button.setChecked(False)
                return
            self.ui.play_button.setToolTip('Stop')
        else:
            self.player.stop()
            self.ui.play_button.setToolTip('Play')

    def on_mute(self, state):
        self.muted = state
        self.set_volume_button_icon(self.ui.volume_button.underMouse() or self.ui.volume_button.hasFocus())
        if self.muted:
            self.player.set_volume(0.0)
            self.ui.volume_button.setToolTip('Unmute')
        else:
            self.player.set_volume(self.ui.volume_slider.value() / 100)
            self.ui.volume_button.setToolTip('Mute')

    def on_volume_change(self, value):
        volume = value / 100
        if not self.muted:
            self.player.set_volume(volume)
        Config.data['volume'] = volume

    def set_widget_icon(self, widget, icon_normal, icon_accent=None, highlight=False):
        if highlight and icon_accent:
            widget.setIcon(icon_accent)
        else:
            widget.setIcon(icon_normal)

    def set_play_button_icon(self, highlight):
        if self.playing:
            self.set_widget_icon(self.ui.play_button, self.stop_icon_normal, self.stop_icon_accent, highlight)
        else:
            self.set_widget_icon(self.ui.play_button, self.play_icon_normal, self.play_icon_accent, highlight)

    def set_volume_button_icon(self, highlight):
        if self.muted:
            self.set_widget_icon(
                self.ui.volume_button,
                self.volume_muted_icon_normal,
                self.volume_muted_icon_accent,
                highlight
            )
        else:
            self.set_widget_icon(
                self.ui.volume_button,
                self.volume_normal_icon_normal,
                self.volume_normal_icon_accent,
                highlight
            )

    def save_settings(self):
        connection_timeout = self.ui.connection_timeout_input.value()
        self.recent_songs_update_worker.timeout = connection_timeout
        Config.data['connection-timeout'] = connection_timeout
        recent_songs_update_time = self.ui.recent_songs_update_time_input.value()
        self.recent_songs_update_worker.update_time = recent_songs_update_time
        Config.data['recent-songs-update-time'] = recent_songs_update_time
        Config.save()

    def update_recent_songs(self, recent_songs):
        for recent_song_widget in self.recent_songs_widgets:
            recent_song_widget.deleteLater()
        self.recent_songs_widgets.clear()
        self.recent_songs_container_layout.removeItem(self.recent_songs_spacer)

        for i, recent_song in enumerate(recent_songs):
            title = recent_song['title']
            timestamp = recent_song['timestamp']
            timestamp_fmt = timeago.format(timestamp)

            recent_song = RecentSong(self.ui.recent_songs_container, title, timestamp_fmt)
            recent_song.setObjectName('recent_song_entry')
            if i % 2 == 0:
                recent_song.setProperty('class', 'alternate')

            self.recent_songs_container_layout.addWidget(recent_song)
            self.recent_songs_widgets.append(recent_song)
        self.recent_songs_container_layout.addSpacerItem(self.recent_songs_spacer)

    def closeEvent(self, event):
        log.debug('Stopping recent songs update worker')
        self.recent_songs_update_worker.stop()
        self.recent_songs_update_worker_thread.quit()
        self.recent_songs_update_worker_thread.wait()
        self.recent_songs_update_worker.deleteLater()
        Config.save()
        super().closeEvent(event)


def gui_main():
    # Enable high DPI support
    if hasattr(QtCore.Qt, 'AA_EnableHighDpiScaling'):
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
    if hasattr(QtCore.Qt, 'AA_UseHighDpiPixmaps'):
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

    app = QtWidgets.QApplication()

    # Set global font size
    font = app.font()
    font.setPointSize(11)
    app.setFont(font)

    # Load and apply stylesheets
    custom_stylesheet = qt_resource(':style.qss')
    qdarktheme_stylesheet = qdarktheme.load_stylesheet('dark')
    app.setStyleSheet(qdarktheme_stylesheet + custom_stylesheet)

    main_window = MainWindow()
    main_window.show()
    return app.exec()



import errno
import json
import logging
import os.path
import platform
import platformdirs
import schema
from pathlib import Path

from .constants import CONFIG_DEFAULTS
from .exceptions import ConfigError


log = logging.getLogger(__name__)


class Config:

    data = {}
    _file = None
    _schema = schema.Schema(
        {
            schema.Optional('output-device'): int,
            schema.Optional('connection-timeout'): float,
            schema.Optional('volume'): float
        },
        ignore_extra_keys=True
    )

    @classmethod
    def init(cls, file=None):
        log.debug('Initializing configuration')

        if file:
            config_dir = Path(os.path.dirname(file))
            config_file = Path(file)
        else:
            if platform.system() == 'Windows':
                dir_name = 'Eternal Radio Player'
            else:
                dir_name = 'eternal-radio-player'
            config_dir = platformdirs.user_config_path(dir_name, False)
            config_file = config_dir.joinpath('config.json')
            log.debug(f"Creating configuration directory: '{config_dir}'")

        try:
            config_dir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            raise ConfigError(f'Could not create config directory: {e}') from e

        log.debug(f'Configuration file: {config_file}')
        cls._file = str(config_file)

    @classmethod
    def load(cls, defaults=None):
        if not defaults:
            defaults = {}

        log.debug('Loading configuration')

        # Load the config file if it exists, else use empty dict
        try:
            with open(cls._file, 'r', encoding='utf-8') as f:
                raw_config_data = json.load(f)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise ConfigError(f'Could not load config: {e}') from e
            raw_config_data = {}
        except json.JSONDecodeError as e:
            raise ConfigError(f'Could not load config: {e}') from e

        # Validate the loaded config and update missing values with defaults
        try:
            cls._schema.validate(raw_config_data)
        except schema.SchemaError as e:
            raise ConfigError(f'Could not load config: {e}') from e

        config_data = CONFIG_DEFAULTS.copy()
        config_data.update({**defaults, **raw_config_data})
        cls.data = config_data

    @classmethod
    def save(cls, defaults=None, write_defaults=False):
        if not defaults:
            defaults = {}

        log.debug('Saving configuration')

        try:
            with open(cls._file, 'w', encoding='utf-8') as f:
                if write_defaults:
                    config_data = CONFIG_DEFAULTS.copy()
                    config_data.update(defaults)
                    json.dump(config_data, f)
                else:
                    json.dump(cls.data, f)
        except (OSError, json.JSONDecodeError) as e:
            raise ConfigError(f'Could not save config: {e}') from e



import json
import logging
import time
import miniaudio
import numpy
import requests
import requests.exceptions
import sounddevice
from threading import Lock

from .constants import (
    PLAYER_FRAME_COUNT,
    RECENT_SONGS_CACHE_TIME,
    RECENT_SONGS_URL,
    REQUEST_TIMEOUT,
    SOUNDDEVICE_HOST_API
)
from .exceptions import PlayerError
from .stream import HTTPStreamSource, stream_request


log = logging.getLogger(__name__)


def get_output_device(device_index=None, host_api_index=SOUNDDEVICE_HOST_API):
    host_api = sounddevice.query_hostapis(host_api_index)
    if device_index is None:
        device_index = host_api['default_output_device']
    elif device_index not in host_api['devices']:
        raise ValueError(f"Output device with index '{device_index}' not found")
    try:
        device = sounddevice.query_devices(device_index, 'output')
    except (sounddevice.PortAudioError, ValueError) as e:
        raise ValueError(f"Output device with index '{device_index}' not found") from e
    return {
        'name': device['name'],
        'sample_rate': int(device['default_samplerate']),
        'index': device_index
    }


def get_output_devices(host_api_index=SOUNDDEVICE_HOST_API):
    host_api = sounddevice.query_hostapis(host_api_index)
    devices = []
    for device_index in host_api['devices']:
        try:
            device = sounddevice.query_devices(device_index, 'output')
        except (sounddevice.PortAudioError, ValueError):
            continue
        devices.append({
            'name': device['name'],
            'sample_rate': int(device['default_samplerate']),
            'index': device_index
        })
    return devices


def check_output_device_params(device_index, channels, sample_rate, sample_width):
    try:
        sounddevice.check_output_settings(
            device=device_index,
            channels=channels,
            dtype=sample_width,
            samplerate=sample_rate
        )
    except sounddevice.PortAudioError as e:
        log.debug(
            f'Incompatible device parameters: Index: {device_index}, Channels: {channels}, '
            f'Width: {sample_width}, Rate: {sample_rate}, Error: {e}'
        )
        return False
    return True


class RadioPlayer:

    recent_songs_cache_time = RECENT_SONGS_CACHE_TIME
    _recent_songs_cache_last = 0.0
    _recent_songs_cache_data = None

    def __init__(self, output_device_index=None, request_timeout=REQUEST_TIMEOUT):
        self.running = False
        self.request_timeout = request_timeout
        try:
            self.output_device = get_output_device(output_device_index)
        except ValueError:
            self.output_device = get_output_device()
        self._stream_source = None
        self._input_stream = None
        self._output_stream = None
        self._linear_volume = 1.0
        self._volume = 1.0
        self._lock = Lock()

    def play(self):
        if self.running:
            log.debug('Cannot start radio player, already running')
            return
        with self._lock:
            log.debug('Starting radio player')
            self._stop_output()
            self._stop_input()
            try:
                self._init()
            except Exception as e:
                raise PlayerError(f'Could not initialize player: {e}') from e
            self._output_stream.start()
            self.running = True

    def stop(self):
        with self._lock:
            log.debug('Stopping radio player')
            self._stop_output()
            self._stop_input()
            self.running = False

    def get_volume(self):
        return self._linear_volume

    def set_volume(self, volume):
        self._linear_volume = volume
        self._volume = volume ** 3

    @classmethod
    def get_recent_songs(cls, timeout=REQUEST_TIMEOUT):
        if time.time() - cls._recent_songs_cache_last <= cls.recent_songs_cache_time:
            return cls._recent_songs_cache_data
        try:
            response = requests.get(url=RECENT_SONGS_URL, timeout=timeout)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise PlayerError(f'Could not fetch recent songs: {e}') from e
        try:
            items = response.json()['items']
            recent_songs = []
            for recent_song in items:
                title = recent_song['title'][:-6].strip() # Remove ID '[XXXX]' at the end
                timestamp = min(int(recent_song['date']), int(time.time()))
                recent_songs.append({'title': title, 'timestamp': timestamp})
        except (json.JSONDecodeError, KeyError, ValueError) as e:
            raise PlayerError('Received bad recent songs data') from e
        cls._recent_songs_cache_data = recent_songs
        cls._recent_songs_cache_last = time.time()
        return recent_songs

    def _init(self):
        request, encoding, sample_rate = stream_request(self.request_timeout)
        if not sample_rate:
            sample_rate = self.output_device['sample_rate']

        for sr in (sample_rate, self.output_device['sample_rate']):
            if check_output_device_params(self.output_device['index'], 2, sr, 'float32'):
                sample_rate = sr
                break
        else:
            raise PlayerError('Could not find supported device parameters')

        log.debug(
            f'Initializing player with request status: {request.status_code}, '
            f'encoding: {encoding}, sample rate: {sample_rate}'
        )
        self._output_stream = sounddevice.OutputStream(
            samplerate=sample_rate,
            blocksize=PLAYER_FRAME_COUNT,
            device=self.output_device['index'],
            channels=2,
            dtype=numpy.float32,
            callback=self._stream_callback_wrapper,
            finished_callback=self._finished_callback
        )
        self._stream_source = HTTPStreamSource(request)
        self._input_stream = miniaudio.stream_any(
            source=self._stream_source,
            source_format=encoding,
            output_format=miniaudio.SampleFormat.FLOAT32,
            nchannels=2,
            sample_rate=sample_rate,
            frames_to_read=PLAYER_FRAME_COUNT,
            dither=miniaudio.DitherMode.TRIANGLE
        )

    def _stream_callback_wrapper(self, *args, **kwargs):
        try:
            self._stream_callback(*args, **kwargs)
            return
        except sounddevice.CallbackAbort:
            raise
        except miniaudio.MiniaudioError as e:
            log.error(f'Decoder error: {e}')
        except Exception as e:
            log.error(f'Stream callback exception: {e}')
        raise sounddevice.CallbackAbort

    def _stream_callback(self, output, frames, _time, status):
        if status and status.output_underflow:
            log.warning('Output buffer underflow')
        raw_samples = self._input_stream.send(frames)
        samples = numpy.asarray(raw_samples, raw_samples.typecode).reshape(-1, 2) * self._volume
        if len(samples) < len(output):
            log.error(f'Length of stream data is less than expected: ({len(samples)}/{len(output)})')
            raise sounddevice.CallbackAbort
        output[:] = samples

    def _finished_callback(self):
        log.debug('Stream callback stopped')
        if self._lock.acquire(False):
            self._stop_input()
            self.running = False
            self._lock.release()

    def _stop_input(self):
        if self._input_stream:
            log.debug('Closing and destroying input stream')
            self._input_stream.close()
            self._input_stream = None

    def _stop_output(self):
        if self._output_stream:
            log.debug('Closing and destroying output stream')
            self._output_stream.abort()
            self._output_stream.close()
            self._output_stream = None
